
#!/bin/sh

##
# Create a new user
#
# Can edit entries in:
# - BSD master.passwd and group files
# - NetInfo domains
#
# $Id$
# $HeadURL$
##
# Copyright (c) 2002 Wilfredo Sanchez Vega, wsanchez@mit.edu.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice, this permission, and the following
# disclaimer notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHORS DISCLAIM ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE
# AUTHORS BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
# WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
##

# Be strict
set -e
set -u

# See <paths.h>
PATH_STDPATH='/usr/bin:/bin:/usr/sbin:/sbin';

##
# Set up PATH
##

PATH="${PATH_STDPATH}";

##
# Functions
##

#
# Version info
#
version ()
{
       file=$(echo '$Id$' |
              cut -d ' ' -f 2);
    version=$(echo '$LastChangedRevision$' |
              cut -d ' ' -f 2);
       date=$(echo '$LastChangedDate$' |
              sed 's|^.*(\(.[^)]*\)).*$|\1|');

    echo "${file} version ${version} (${date})";
    echo "Copyright (c) 2002 Wilfredo Sanchez Vega.";
}

#
# Command line usage information
#
usage ()
{
    echo "Usage: $(basename "${Program}") [-Viv] [-D domain] [-n name] [-m uid:gid]";
    echo "	[-u user[:id]] [-g group[:id]] [-G group[,...]] [-c class]";
    echo "	[-H dir] [-h home] [-s shell]";
    echo "Options:";
    echo "	-h             Print this help and exit";
    echo "	-V             Print version and exit";
    echo "	-i             Interactive mode, prompt for information";
    echo "	-v             Be verbose";
    echo "	-D domain      NetInfo domain to edit (instead of flat files)";
    echo "	-n name        User's real name";
    echo "	-m uid:gid     Minimum user and group ID when generating new IDs [${MinimumUID}:${MinimumGID}]";
    echo "	-u user        User's login name";
    echo "	-u user:id     User's login name and ID";
    echo "	-g group       User's login group name [same as login name]";
    echo "	-g group:id    User's login group name and ID";
    echo "	-G group,...   User's group memberships [${StaffGroup}]";
    echo "	-c class       User's login class [${LoginClass}]";
    echo "	-H dir         Directory to place user's home directory into (overrides -h) [${HomePartition}]";
    echo "	-d home        User's home directory (overrides -H)";
    echo "	-s shell       User's login shell [${LoginShell}]";
    if [ "${1-}" == "-" ]; then return 0; fi;
    exit 64;
}

#
# oops <message>
# Print error message and exit
#
oops ()
{
    local message="$*";

    echo "$(basename "${Program}"): $*";
    exit 64;
}

#
# split_arg <argument> <var1> <match1> [<var2> <match2>]
# Parse <x>:<y> type command line arguments
# Values are stored into specified variables if they match the
#  given regex patterns.
#
split_arg ()
{
    local arg="$1";
    local   A="$2";
    local  mA="$3"; shift 3;
    local   B="";
    local  mB="";

    if [ $# != 0 ]; then B="$1"; mB="$2"; fi;

    if [ -z "${B}" ]; then
        cmd=$(echo "${arg}" | sed -e 's|^\('"${mA}"'\)$|'${A}'=\1;|')
    else
        cmd=$(echo "${arg}" | sed -e 's|^\('"${mA}"'\):\('${mB}'\)$|'${A}'='\''\1'\'';'${B}'='\''\2'\'';|')
    fi;

    if [ "${cmd}" = "${arg}" ]; then return 1; fi;

    eval "${cmd}";

    return 0;
}

#
# prompt <var> <match>
# Interactively prompt for information.
# Response must match the given regex pattern, and is stored
#  in the specified variable.
#
prompt ()
{
    local prompt="$1";
    local    var="$2";
    local  match="$3";

    local val; eval 'val="$'${var}'"';
    local response="";

    while [ -z "${response}" ]; do
        echo -n "Enter ${prompt}";
        if [ -n "${val}" ]; then echo -n " [${val}]"; fi;
        echo -n ": "; read response;

        if [ -z "${response}" ]; then response="${val}"; fi;

        local test=$(echo ${response} | sed -e 's|^\('"${match}"'\)$|\1x|');

        if [ "${test}" = "${response}x" ]; then
            # Quote "'" characters to prevent eval from going amock
            local cmd=\'$(echo "${response}" | sed -e "s|'|'\\\\''|g")\';

            eval "${var}=${cmd}";
            response="x";
        else
            echo "Invalid ${prompt}.";
            response="";
        fi;

    done;
}

#
# boolean_prompt <prompt>
# Interactively prompt for a yes/no answer.
#
boolean_prompt ()
{
    local prompt;
    local response;

    prompt="$1";

    while (true); do
        echo -n "${prompt} "; read response;

        case "${response}" in
          y|Y|yes|YES|Yes) return 0; ;;
          n|N|no |NO |No ) return 1; ;;
        esac;
    done;
}

#
# verify_nidomain <domain>
# Verify NetInfo domain
#
verify_nidomain ()
{
    local domain="$1";

    if nicl "${domain}" -echo "" > /dev/null 2>&1; then
        return 0;
    else
        return 1;
    fi;
}

#
# user_id <name>
# Get ID for a user
#
user_id ()
{
    local name="$1";
    local id;

    if [ -n "${NetInfoDomain}" ]; then
        id=$(nicl "${NetInfoDomain}" -read "/users/${name}" uid 2> /dev/null |
             sed 's|^uid: ||');
    else
        id=$(grep "^${name}:" "${PasswordFile}" |
             sed 's|^[^:]*:[^:]*:\([^:]*\):.*$|\1|');
    fi;

    if [ -z "${id}" ]; then return 1; fi;

    echo "${id}";

    return 0;
}

#
# user_name <id>
# Get name for a user
user_name ()
{
    local id="$1";
    local name;

    if [ -n "${NetInfoDomain}" ]; then
        name=$(nigrep "${id}" "${NetInfoDomain}" |
               grep '^[0-9][0-9]* /users/[^:]*: *uid '"${id}"'$' |
               sed 's|^[0-9][0-9]* /users/\([^:]*\):.*|\1|');
    else
        name=$(grep '^[^:]*:[^:]*:'"${id}"':' "${PasswordFile}" |
               sed 's|^\([^:]*\):.*|\1|');
    fi;

    if [ -z "${name}" ]; then return 1; fi;

    echo "${name}";

    return 0;
}

#
# group_id <name>
# Get ID for a group
#
group_id ()
{
    local name="$1";
    local id;

    if [ -n "${NetInfoDomain}" ]; then
        id=$(nicl "${NetInfoDomain}" -read "/groups/${name}" gid 2> /dev/null |
             sed 's|^gid: ||');
    else
        id=$(grep "^${name}:" "${GroupFile}" |
             sed 's|^[^:]*:[^:]*:\([^:]*\):.*$|\1|');
    fi;

    if [ -z "${id}" ]; then return 1; fi;

    echo "${id}";

    return 0;
}

#
# group_name <id>
# Get name for a group
#
group_name ()
{
    local id="$1";
    local name;

    if [ -n "${NetInfoDomain}" ]; then
        name=$(nigrep "${id}" "${NetInfoDomain}" |
               grep '^[0-9][0-9]* /groups/[^:]*: *gid '"${id}"'$' |
               sed 's|^[0-9][0-9]* /groups/\([^:]*\):.*|\1|');
    else
        name=$(grep '^[^:]*:[^:]*:'"${id}"':' "${GroupFile}" |
               sed 's|^\([^:]*\):.*|\1|');
    fi;

    if [ -z "${name}" ]; then return 1; fi;

    echo "${name}";

    return 0;
}

#
# find_userid <start>
# Find an avaliable user ID
#
find_userid ()
{
    local id="$1";

    while user_name "${id}" > /dev/null; do
        id=$((${id} + 1));
    done;

    echo "${id}";
}

#
# find_groupid <start>
# Find an avaliable group ID
#
find_groupid ()
{
    local id="$1";

    while group_name "${id}" > /dev/null; do
        id=$((${id} + 1));
    done;

    echo "${id}";
}

##
# Handle command line
##

# Constants
       Program="$0";
  PasswordFile="/etc/master.passwd";
     GroupFile="/etc/group";
    StaffGroup="staff";
     Skeletons="/usr/share/skel /etc/skel";

# User variables
  Interactive="false";
      Verbose="false";
NetInfoDomain="";
   MinimumUID="1000";
   MinimumGID="1000";
     RealName="";
     UserName="";
       UserID="";
    GroupName="";
      GroupID="";
    GroupList="";
   LoginClass="";
       Change="0";
       Expire="0";
HomePartition="/Users";
HomeDirectory="";
   LoginShell="/bin/tcsh";

while getopts 'hVivD:n:m:u:g:G:c:H:d:s:' option; do
  case "$option" in
    '?') usage; ;;
    'h') version; echo ""; usage -; exit 0; ;;
    'V') version; exit 0; ;;

    'i')    Interactive="true"; ;;
    'v')        Verbose="true"; ;;
    'D')  NetInfoDomain="${OPTARG}"; ;;
    'n')       RealName="${OPTARG}"; ;;
    'G')      GroupList="${OPTARG}"; ;;
    'c')     LoginClass="${OPTARG}"; ;;
    's')     LoginShell="${OPTARG}"; ;;

    'H') HomePartition="${OPTARG}"; HomeDirectory=""; ;;
    'd') HomeDirectory="${OPTARG}"; HomePartition=""; ;;

    'm')
      if ! split_arg "${OPTARG}" MinimumUID '[0-9][0-9]*' MinimumGID '[0-9][0-9]*'; then
          oops "Invalid minimum IDs: ${OPTARG}";
      fi;
      ;;
    'u')
      if ! split_arg "${OPTARG}" UserName '[a-z_][a-z_0-9]*' UserID '[0-9]*' &&
         ! split_arg "${OPTARG}" UserName '[a-z_][a-z_0-9]*'; then
          oops "Invalid user name/ID: ${OPTARG}";
      fi;
      ;;
    'g')
      if ! split_arg "${OPTARG}" GroupName '[a-z_][a-z_0-9]*' GroupID '[0-9]*' &&
         ! split_arg "${OPTARG}" GroupName '[a-z_][a-z_0-9]*'; then
          oops "Invalid group name/ID: ${OPTARG}";
      fi;
      ;;
  esac;
done;
shift $((${OPTIND} - 1));

if [ $# != 0 ]; then oops "Unrecognized arguments:" "$@"; fi;

##
# Interact and verify values
##

if [ -n "${NetInfoDomain}" ] && ! verify_nidomain "${NetInfoDomain}"; then
    echo "Unable to connect to NetInfo domain: ${NetInfoDomain}";
    exit 1;
fi;

if "${Interactive}"; then
    Verbose="true";
fi;

if "${Interactive}"; then
    prompt "real name" RealName '.*';
fi;

if "${Interactive}"; then
    fallback="${UserName}";
    while true; do
        prompt "user name" UserName '[a-z_][a-z_0-9]*';

        if user_id "${UserName}" > /dev/null; then
            echo "User name ${UserName} is already in use.";
            UserName="${fallback}";
        else
            break;
        fi;
    done;
else
    if user_id "${UserName}" > /dev/null; then
        echo "User name ${UserName} is already in use.";
        exit 1;
    fi;
fi;

if [ -z "${UserName}" ]; then oops "No user name."; fi;

if [ -z "${UserID}" ]; then UserID=$(find_userid "${MinimumUID}"); fi;

if "${Interactive}"; then
    fallback="${UserID}";
    while true; do
        prompt "user ID" UserID '[0-9][0-9]*';

        if user_name "${UserID}" > /dev/null; then
            echo "User ID ${UserID} is already in use.";
            UserID="${fallback}";
        else
            break;
        fi;
    done;
else
    if name=$(user_name "${UserID}"); then
        echo "User ID ${UserID} (${name}) is already in use.";
        exit 1;
    fi;
fi;

if [ -z "${UserID}" ]; then oops "No user ID."; fi;

if [ -z "${GroupName}" ] && ! group_id "${UserName}" > /dev/null; then
    GroupName="${UserName}";
fi;

if "${Interactive}"; then
    prompt "login group" GroupName '[a-z_][a-z_0-9]*';
fi;

if [ -z "${GroupName}" ]; then oops "No login group name."; fi;

if [ -z "${GroupID}" ]; then GroupID=$(group_id "${GroupName}") || true; fi;
if [ -z "${GroupID}" ]; then GroupID=$(find_groupid "${MinimumGID}"); fi;

if "${Interactive}"; then
    prompt "login group ID" GroupID '[0-9][0-9]*';
fi;

# FIXME: verify that if GroupName exists and GroupID is set, that GroupID matches GroupName
# FIXME: verify group ID is available

if [ -z "${GroupID}" ]; then oops "No login group ID."; fi;

if "${Interactive}"; then
    echo "Add user to groups here. Enter empty group to finish list.";

    while true; do
        if [ -n "${GroupList}" ]; then echo "Groups: ${GroupList}"; fi;
        newgroup="";
        prompt "secondary group" newgroup '\([a-z_][a-z_0-9]*\)*';
        if [ -n "${newgroup}" ]; then
            if ! group_id "${newgroup}" > /dev/null; then
                echo "Group ${newgroup} doesn't exist in this domain.";
                echo "However, it's possible that this group is still valid on the system."
                if ! boolean_prompt "Would you like to add ${newgroup} to the group list anyway?"; then
                    continue;
                fi;
            fi;
            if [ -z "${GroupList}" ]; then
                GroupList="${newgroup}";
            else
                GroupList="${GroupList},${newgroup}";
            fi;
        else
            break;
        fi;
    done;
fi;

if "${Interactive}"; then
    prompt "login class" LoginClass '\([a-z][a-z_0-9]*\)*';
fi;

# if [ -n "${LoginClass}" ]; then
#     # FIXME verify login class
#     # eg. check that it exists in login.access
# fi;

if [ -z "${HomeDirectory}" ] && [ -n "${HomePartition}" ]; then
    HomeDirectory="${HomePartition}/${UserName}";
fi;

if "${Interactive}"; then
    prompt "home directory" HomeDirectory '.*';
fi;

if [ -z "${HomeDirectory}" ]; then oops "No home directory."; fi;

if "${Interactive}"; then
    prompt "login shell" LoginShell '.*';
fi;

if [ -z "${LoginShell}" ]; then oops "No login shell."; fi;

if [ ! -x "${LoginShell}" ]; then oops "Login shell is not executable."; fi;

if "${Verbose}"; then
    echo "User information:";
    echo "";
    echo "NetInfoDomain=${NetInfoDomain}";
    echo "     RealName=${RealName}";
    echo "   MinimumUID=${MinimumUID}";
    echo "   MinimumGID=${MinimumGID}";
    echo "     UserName=${UserName}";
    echo "       UserID=${UserID}";
    echo "    GroupName=${GroupName}";
    echo "      GroupID=${GroupID}";
    echo "    GroupList=${GroupList}";
    echo "   LoginClass=${LoginClass}";
    echo "       Change=${Change}";
    echo "       Expire=${Expire}";
    echo "HomePartition=${HomePartition}";
    echo "HomeDirectory=${HomeDirectory}";
    echo "   LoginShell=${LoginShell}";
    echo "";
fi;
# exit 1;

##
# Do The Right Thing
##

if [ -n "${NetInfoDomain}" ]; then

    if "${Verbose}"; then
        echo "Creating new user ${UserName} in NetInfo domain ${NetInfoDomain}...";
    fi;
    nicl "${NetInfoDomain}" > /dev/null <<EOF
create "/users/${UserName}"
create "/users/${UserName}" uid "${UserID}"
create "/users/${UserName}" gid "${GroupID}"
create "/users/${UserName}" passwd "*"
create "/users/${UserName}" class "${LoginClass}"
create "/users/${UserName}" change "${Change}"
create "/users/${UserName}" expire "${Expire}"
create "/users/${UserName}" realname "${RealName}"
create "/users/${UserName}" home "${HomeDirectory}"
create "/users/${UserName}" shell "${LoginShell}"
create "/users/${UserName}" _writers_passwd "${UserName}"
create "/users/${UserName}" _writers_realname "${UserName}"
EOF

    if ! group_id "${GroupName}" > /dev/null; then
        if "${Verbose}"; then
            echo "Creating new group ${GroupName} in NetInfo domain ${NetInfoDomain}...";
        fi;
        nicl "${NetInfoDomain}" > /dev/null <<EOF
create "/groups/${GroupName}"
create "/groups/${GroupName}" gid "${GroupID}"
create "/groups/${GroupName}" users
create "/groups/${GroupName}" passwd "*"
EOF
    fi;

    ifs="${IFS}"; IFS=",";
    for group in ${GroupList}; do
        if "${Verbose}"; then
            echo "Adding user ${UserName} to group ${group} in NetInfo domain ${NetInfoDomain}...";
        fi;
        nicl "${NetInfoDomain}" > /dev/null <<EOF
append "/groups/${group}" users "${UserName}"
EOF
    done;
    IFS="${ifs}";

else

    if "${Verbose}"; then
        echo "Creating new user ${UserName} in ${PasswordFile}...";
    fi;
    tmp=$(mktemp "${PasswordFile}.XXXXXX");
    cp "${PasswordFile}" "${tmp}";
    echo "${UserName}:*:${UserID}:${GroupID}:${LoginClass}:${Change}:${Expire}:${RealName}:${HomeDirectory}:${LoginShell}" >> "${tmp}";
    pwd_mkdb -p "${tmp}";
    rm -f "${tmp}";

    if ! group_id "${GroupName}"; then
        if "${Verbose}"; then
            echo "Creating new group ${GroupName} in ${GroupFile}...";
        fi;
        echo "${GroupName}:*:${GroupID}:" >> "${GroupFile}";
    fi;

    ifs="${IFS}"; IFS=",";
    for group in ${GroupList}; do
        if "${Verbose}"; then
            echo "Adding user ${UserName} to group ${group} in ${GroupFile}...";
        fi;
        tmp=$(mktemp -t group);
        cat "${GroupFile}" \
            | sed -e 's|^\('"${group}"':.*[^:]\)$|\1',"${UserName}"'|' \
            | sed -e 's|^\('"${group}"':.*:\)$|\1'"${UserName}"'|' \
            >> "${tmp}";
        cp "${tmp}" "${GroupFile}";
        rm -f "${tmp}";
    done;
    IFS="${ifs}";

fi;

if "${Verbose}"; then
    echo "Creating user home directory ${HomeDirectory}...";
fi;
install -d -m 755 -o "${UserID}" -g "${GroupID}" "${HomeDirectory}";

for dir in ${Skeletons}; do
    if [ -d "${dir}" ]; then
        if "${Verbose}"; then
            echo "Copying files from ${dir} to user home directory...";
        fi;
        cd "${dir}" && cp -Rp . "${HomeDirectory}";
    fi;
    chown -R "${UserID}":"${GroupID}" "${HomeDirectory}";
done;
